package team.frc.morpheus.robot.commands.drivebase;

import edu.wpi.first.wpilibj2.command.CommandBase;
import team.frc.morpheus.robot.Robot;
import team.frc.morpheus.robot.Constants;
import team.frc.morpheus.robot.subsystems.DriveBase;
import team.frc.morpheus.util.input.Coordinate;
import team.frc.morpheus.util.input.Filters;
import team.frc.morpheus.vision.limelight.DoubleTargetTrackerLimelight;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.function.BooleanSupplier;
import java.util.function.DoubleSupplier;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 *
 */
public class DriverControl extends CommandBase {
    @FunctionalInterface
    public interface Feature extends Function<Coordinate, Coordinate> {}

    private final Supplier<Coordinate> coordinateSupplier;
    protected final DriveBase driveBase;
    private final LinkedList<Feature> features;

    public DriverControl(Supplier<Coordinate> coordinateSupplier) {
        this.coordinateSupplier = coordinateSupplier;
        driveBase = DriveBase.getInstance();
        hasRequirement(driveBase);
        features = new LinkedList<>();
    }

    // returns itself, so you can chain away!
    public DriverControl withFeature(Feature feat) {
        features.add(feat);
        return this;
    }

    public DriverControl withFeatures(Feature... feat) {
        features.addAll(Arrays.asList(feat));
        return this;
    }

    // Called just before this Command runs the first time
    public void initialize() {
    }

    // Called repeatedly when this Command is scheduled to run
    public void execute() {
        Coordinate input = coordinateSupplier.get();
        for (Feature feature : features) input = feature.apply(input);
        driveBase.set(input);

        //Robot.driveBase.setCheesy(Filters.threshold(Robot.m_oi.driver.getLeftStick(), 0.1), Robot.m_oi.driver.getButtonStart());
    }

    // Called once after isFinished returns true
    protected void end() {
    }

    // Called when another command which requires one or more of the same
    // subsystems is scheduled to run
    protected void interrupted() {
        driveBase.set(new Coordinate.Cartesian(0, 0));
    }

    public static class Features {

        public static Feature buildAssisted(DoubleSupplier correctionExtentSupplier, DoubleSupplier lrAdjustmentSupplier) {
            return (Coordinate in) -> {
                double correctionExtent = correctionExtentSupplier.getAsDouble();
                double lrAdjustment = lrAdjustmentSupplier.getAsDouble();
                return new Coordinate.Cartesian(in.x() + correctionExtent * lrAdjustment * Constants.ASSISTED_DRIVING_KP, in.y());
            };
        }

        public static Feature buildSlowMode(BooleanSupplier enabled, final double scaleX, final double scaleY) {
            return (Coordinate in) -> {
                if (enabled.getAsBoolean()){
                    return in;
                } else {
                    return new Coordinate.Cartesian(in.x() * scaleX, in.y() * scaleY);
                }
            };
        }
    }
}
