package team.frc.morpheus.robot.subsystems;

import com.ctre.phoenix.motorcontrol.NeutralMode;
import com.ctre.phoenix.motorcontrol.TalonFXControlMode;
import com.ctre.phoenix.motorcontrol.TalonFXFeedbackDevice;
import com.ctre.phoenix.motorcontrol.can.TalonFX;
import com.kauailabs.navx.frc.AHRS;
import edu.wpi.first.wpilibj.SPI;
import edu.wpi.first.wpilibj2.command.Subsystem;
import jaci.pathfinder.Pathfinder;
import jaci.pathfinder.Trajectory;
import jaci.pathfinder.followers.EncoderFollower;
import jaci.pathfinder.modifiers.TankModifier;
import team.frc.morpheus.robot.Constants;
import team.frc.morpheus.util.input.Coordinate;
import team.frc.morpheus.util.input.Filters;

import java.util.Optional;
import java.util.Timer;
import java.util.TimerTask;
import java.util.function.Function;

/**
 *
 */
public class DriveBase implements Subsystem {
    public static final double WIDTH = 18.25; // Inches
    // dt in seconds, max_velocity in m/s, max_acceleration in m/s^2, max_jerk in m/s^3
    public static final Trajectory.Config CONFIG = new Trajectory.Config(Trajectory.FitMethod.HERMITE_CUBIC,
            Trajectory.Config.SAMPLES_HIGH, 0.05, 3.48, 3.48, 60); // max_jerk is unlikely to be correct
    private static final double TURN_SENSITIVITY = 1, KP_GYRO = 0.007; //Entropy = 0.007
    private double qsAccumulator = 0;
    private final TalonFX left, leftFollower, right, rightFollower;
    private AHRS navX;
    private Timer trajectoryScheduler;
    private Optional<TimerTask> trajectoryRunner = Optional.empty();
    private static Optional<DriveBase> instance = Optional.empty();
    private final int ticksPerRev = 16674; // Rounded from 16673.62
    private final double wheelDiameter = 0.1016; // Meters
    private boolean breaksEnabled;

    private DriveBase() {
        instance = Optional.of(this);
        left = new TalonFX(Constants.DriveBase.LEFT_DRIVE);
        right = new TalonFX(Constants.DriveBase.RIGHT_DRIVE);
        leftFollower = new TalonFX(Constants.DriveBase.LEFT_FOLLOWER);
        rightFollower = new TalonFX(Constants.DriveBase.RIGHT_FOLLOWER);
        breaksEnabled = false;

        try {
            navX = new AHRS(SPI.Port.kMXP);
        } catch (RuntimeException e) {
            System.out.println("Error instantiating navX MXP:  " + e.getMessage());
        }

        trajectoryScheduler = new Timer();

        left.setInverted(!Constants.INTAKE_SIDE_FORWARD);
        right.setInverted(Constants.INTAKE_SIDE_FORWARD);

//        leftFollower.follow(left);
//        rightFollower.follow(right);

        leftFollower.setInverted(!Constants.INTAKE_SIDE_FORWARD);
        rightFollower.setInverted(Constants.INTAKE_SIDE_FORWARD);

        left.configSelectedFeedbackSensor(TalonFXFeedbackDevice.IntegratedSensor, 0, 0);
        right.configSelectedFeedbackSensor(TalonFXFeedbackDevice.IntegratedSensor, 0, 0);

        // TODO: apply pid settings to each controller via getPIDController()

        // /* set closed loop gains in slot */
        // left.selectProfileSlot(0, 0);
        // left.config_kF(0, 1.4, 0);
        // left.config_kP(0, 16, 0);
        // left.config_kI(0, 0, 0);
        // left.config_kD(0, 0, 0);
        // /* zero the sensor */
        // left.setSelectedSensorPosition(0, 0, 0);

        // /* set closed loop gains in slot */
        // right.selectProfileSlot(0, 0);
        // right.config_kF(0, 1.65, 0);
        // right.config_kP(0, 16, 0);
        // right.config_kI(0, 0, 0);
        // right.config_kD(0, 0, 0);
        // /* zero the sensor */
        // right.setSelectedSensorPosition(0, 0, 0);

    }

    public static synchronized DriveBase getInstance() {
        return instance.orElseGet(team.frc.morpheus.robot.subsystems.DriveBase::new);
    }

    public void set(Coordinate pt) {
        cancelTrajectory();
        Filters.standardScaled(pt, 0.1);
        // left.setParameter(ConfigParameter.kCtrlType, ControlType.kDutyCycle.value);
        // right.setParameter(ConfigParameter.kCtrlType, ControlType.kDutyCycle.value);

        directSet(pt.y() + pt.x(), pt.y() - pt.x());
        // left.set(pt.y() + pt.x());
        // right.set(pt.y() - pt.x());
    }

    public void setCubic(Coordinate pt) {
        cancelTrajectory();
        Filters.standardScaled(pt, 0.1);
        directSet(Math.pow(pt.y() + pt.x(), 3), Math.pow(pt.y() - pt.x(), 3));
    }

    // inspired by
    // https://github.com/Team254/FRC-2016-Public/blob/master/src/com/team254/frc2016/CheesyDriveHelper.java
    public void setCheesy(Coordinate pt, boolean quickTurn) {
        // TRAJ cancelTrajectory();
        Filters.constrain(pt);

        double overPower;
        double angularPower;

        if (quickTurn) {
            if (Math.abs(pt.y()) < 0.2) {
                double alpha = 0.1;
                qsAccumulator = (1 - alpha) * qsAccumulator + alpha * pt.x() * 2;
            }
            overPower = 1.0;
            angularPower = pt.x();
        } else {
            overPower = 0.0;
            angularPower = Math.abs(pt.y()) * pt.x() * TURN_SENSITIVITY - qsAccumulator;
            if (qsAccumulator > 1)
                qsAccumulator -= 1;
            else if (qsAccumulator < -1)
                qsAccumulator += 1;
            else
                qsAccumulator = 0;
        }

        double rightOut = pt.y() - angularPower;
        double leftOut = pt.y() + angularPower;

        if (leftOut > 1) {
            rightOut -= overPower * (leftOut - 1);
            leftOut = 1;
        } else if (rightOut > 1) {
            leftOut -= overPower * (rightOut - 1);
            rightOut = 1;
        } else if (leftOut < -1) {
            rightOut += overPower * (-1 - leftOut);
            leftOut = -1;
        } else if (rightOut < -1) {
            leftOut += overPower * (-1 - rightOut);
            rightOut = -1;
        }

        directSet(leftOut, rightOut);
    }

    public void directSet(double l, double r) {
        if (Constants.INTAKE_SIDE_FORWARD) {
            left.set(TalonFXControlMode.PercentOutput, r);
            leftFollower.set(TalonFXControlMode.PercentOutput, r); // disable if following is re-enabled
            right.set(TalonFXControlMode.PercentOutput, l);
            rightFollower.set(TalonFXControlMode.PercentOutput, l); // disable if following is re-enabled
        } else {
            left.set(TalonFXControlMode.PercentOutput, l);
            leftFollower.set(TalonFXControlMode.PercentOutput, l); // disable if following is re-enabled
            right.set(TalonFXControlMode.PercentOutput, r);
            rightFollower.set(TalonFXControlMode.PercentOutput, r); // disable if following is re-enabled
        }
    }

    public int getLeftPosition() {
        // return left.getEncoder().getWristPosition();
        if (!Constants.INTAKE_SIDE_FORWARD)
            return left.getSelectedSensorPosition(0);
        else
            return right.getSelectedSensorPosition(0);
    }

    public int getRightPosition() {
        // return right.getEncoder().getWristPosition();
        if (!Constants.INTAKE_SIDE_FORWARD)
            return right.getSelectedSensorPosition(0);
        else
            return left.getSelectedSensorPosition(0);
    }


    public void enableBrakes() {
        left.setNeutralMode(NeutralMode.Brake);
        right.setNeutralMode(NeutralMode.Brake);
        leftFollower.setNeutralMode(NeutralMode.Brake);
        rightFollower.setNeutralMode(NeutralMode.Brake);
        breaksEnabled = true;
    }

    public void disableBrakes() {
        left.setNeutralMode(NeutralMode.Coast);
        right.setNeutralMode(NeutralMode.Coast);
        leftFollower.setNeutralMode(NeutralMode.Coast);
        rightFollower.setNeutralMode(NeutralMode.Coast);
        breaksEnabled = false;
    }

    public boolean areBrakesEnabled() {
        return breaksEnabled;
    }

    public boolean isMoving() {
        return Math.abs(left.getSelectedSensorVelocity()) + Math.abs(right.getSelectedSensorVelocity()) != 0;
    }

    public void executeTrajectory(Trajectory trajectory, boolean reverse) {
        executeTrajectory(trajectory, reverse, false, Optional.empty());
    }

    public void executeTrajectory(Trajectory trajectory, boolean reverse, boolean useGyro, Optional<Function<Double, Double>> additionalCorrection) {
        cancelTrajectory();

        TankModifier modifier = new TankModifier(trajectory).modify(WIDTH * 0.0254);
        EncoderFollower leftEF = new EncoderFollower(modifier.getLeftTrajectory());
        EncoderFollower rightEF = new EncoderFollower(modifier.getRightTrajectory());

        if (reverse) {
            leftEF.configureEncoder(-getRightPosition(), ticksPerRev, wheelDiameter);
            rightEF.configureEncoder(-getLeftPosition(), ticksPerRev, wheelDiameter);
            leftEF.configurePIDVA(0.5, 0, 0, 0.287, 0); // kv was .075
            rightEF.configurePIDVA(0.5, 0, 0, 0.287, 0); // kv was .075
        } else {
            leftEF.configureEncoder(getLeftPosition(), ticksPerRev, wheelDiameter);
            rightEF.configureEncoder(getRightPosition(), ticksPerRev, wheelDiameter);
            leftEF.configurePIDVA(0.5, 0, 0, 0.287, 0);
            rightEF.configurePIDVA(0.5, 0, 0, 0.287, 0);
        }

        if (useGyro) {
            resetGyroAngle();
        }

        final int[] i = {0};

        trajectoryRunner = Optional.of(new TimerTask() {
            @Override
            public void run() {
                if (!leftEF.isFinished() || !rightEF.isFinished()) {
                    if (reverse) {
                        double l = leftEF.calculate(-getRightPosition());
                        double r = rightEF.calculate(-getLeftPosition());
                        if (useGyro) {
                            double gyroCorrection = KP_GYRO
                                    * Pathfinder.boundHalfDegrees((leftEF.getHeading() * 180.0 / Math.PI) - getGyroAngle());
                            l -= gyroCorrection; // signs may need to be flipped in reverse case...
                            r += gyroCorrection;
                        }
                        if (additionalCorrection.isPresent()) {
                            final double ac = additionalCorrection.get().apply(((double) i[0]++) / trajectory.length());
                            l -= ac;
                            r += ac;
                        }
                        if (l > 1 || r > 1) {
                            double d = Math.max(l, r);
                            l = l / d;
                            r = r / d;
                        }
                        directSet(-r, -l);
                    } else {
                        double l = leftEF.calculate(getLeftPosition());
                        double r = rightEF.calculate(getRightPosition());
                        if (useGyro) {
                            double gyroCorrection = KP_GYRO
                                    * Pathfinder.boundHalfDegrees((leftEF.getHeading() * 180.0 / Math.PI) - getGyroAngle());
                            l -= gyroCorrection;
                            r += gyroCorrection;
                        }
                        if (additionalCorrection.isPresent()) {
                            final double ac = additionalCorrection.get().apply(((double) i[0]++) / trajectory.length());
                            l += ac;
                            r -= ac;
                        }
                        if (l > 1 || r > 1) {
                            double d = Math.max(l, r);
                            l = l / d;
                            r = r / d;
                        }
                        directSet(l, r);
                    }
                } else {
                    cancelTrajectory();
                }
            }
        });
        trajectoryScheduler.scheduleAtFixedRate(trajectoryRunner.get(), 0, 50);
    }

    public synchronized void cancelTrajectory() {
        if (trajectoryRunner.isPresent()) {
            trajectoryRunner.get().cancel();
            trajectoryRunner = Optional.empty();
            directSet(0, 0); // better to be safe than sorry
        }
    }

    public boolean isTrajectoryInProgress() {
        return trajectoryRunner.isPresent();
    }

    public double getGyroAngle() {
        return -navX.getYaw(); // sign changed without changing the position of the navx!!!
    }

    public void resetGyroAngle() {
        navX.zeroYaw();
    }

}
